// auto-generated DO NOT EDIT

#pragma once

#include <algorithm>
#include <array>
#include <functional>
#include <mutex>
#include <rclcpp/node.hpp>
#include <rclcpp_lifecycle/lifecycle_node.hpp>
#include <set>
#include <sstream>
#include <string>
#include <type_traits>
#include <vector>

#include <fmt/core.h>
#include <fmt/format.h>
#include <fmt/ranges.h>

#include <parameter_traits/parameter_traits.hpp>

#include "example_validators.hpp"


namespace admittance_controller {
    struct Params {
        double scientific_notation_num = 1e-11;
        std::string interpolation_mode = "spline";
        std::vector<std::string> joints = {"joint1", "joint2", "joint3"};
        parameter_traits::FixedSizeString<25> fixed_string = {"string_value"};
        parameter_traits::FixedSizeArray<double, 10> fixed_array = {{1.0, 2.3, 4.0, 5.4, 3.3}};
        parameter_traits::FixedSizeString<25> fixed_string_no_default;
        std::vector<std::string> command_interfaces;
        std::vector<std::string> state_interfaces;
        std::vector<std::string> chainable_command_interfaces;
        bool enable_parameter_update_without_reactivation = true;
        bool use_feedforward_commanded_input = false;
        struct Pid {
            double rate = 0.005;
            struct MapJoints {
                double p = 1.0;
                double i;
                double d = 1.0;
            };
            std::map<std::string, MapJoints> joints_map;
        } pid;
        struct Kinematics {
            std::string plugin_name;
            std::string plugin_package;
            std::string base;
            std::string tip;
            double alpha = 0.0005;
            std::string group_name;
        } kinematics;
        struct FtSensor {
            std::string name;
            double filter_coefficient = 0.005;
            struct Frame {
                std::string id;
                bool external = false;
            } frame;
        } ft_sensor;
        struct Control {
            struct Frame {
                std::string id;
                bool external = false;
            } frame;
        } control;
        struct FixedWorldFrame {
            struct Frame {
                std::string id;
                bool external = false;
            } frame;
        } fixed_world_frame;
        struct GravityCompensation {
            struct Frame {
                std::string id;
                bool external = false;
            } frame;
            struct Cog {
                std::vector<double> pos;
                double force = std::numeric_limits<double>::quiet_NaN();
            } CoG;
        } gravity_compensation;
        struct Admittance {
            std::vector<bool> selected_axes;
            std::vector<double> mass;
            std::vector<double> damping_ratio;
            std::vector<double> stiffness;
        } admittance;
        // for detecting if the parameter struct has been updated
        rclcpp::Time __stamp;
    };
    struct StackParams {
        double scientific_notation_num = 1e-11;
        parameter_traits::FixedSizeString<25> fixed_string = {"string_value"};
        parameter_traits::FixedSizeArray<double, 10> fixed_array = {{1.0, 2.3, 4.0, 5.4, 3.3}};
        parameter_traits::FixedSizeString<25> fixed_string_no_default;
        bool enable_parameter_update_without_reactivation = true;
        bool use_feedforward_commanded_input = false;
        struct Pid {
            double rate = 0.005;
        } pid;
        struct Kinematics {
            double alpha = 0.0005;
        } kinematics;
        struct FtSensor {
            double filter_coefficient = 0.005;
            struct Frame {
                bool external = false;
            } frame;
        } ft_sensor;
        struct Control {
            struct Frame {
                bool external = false;
            } frame;
        } control;
        struct FixedWorldFrame {
            struct Frame {
                bool external = false;
            } frame;
        } fixed_world_frame;
        struct GravityCompensation {
            struct Frame {
                bool external = false;
            } frame;
            struct Cog {
                double force = std::numeric_limits<double>::quiet_NaN();
            } CoG;
        } gravity_compensation;
    };

  class ParamListener{
  public:
    // throws rclcpp::exceptions::InvalidParameterValueException on initialization if invalid parameter are loaded
    ParamListener(rclcpp::Node::SharedPtr node)
    : ParamListener(node->get_node_parameters_interface()) {}

    ParamListener(rclcpp_lifecycle::LifecycleNode::SharedPtr node)
    : ParamListener(node->get_node_parameters_interface()) {}

    ParamListener(const std::shared_ptr<rclcpp::node_interfaces::NodeParametersInterface>& parameters_interface){
      parameters_interface_ = parameters_interface;
      declare_params();
      auto update_param_cb = [this](const std::vector<rclcpp::Parameter> &parameters){return this->update(parameters);};
      handle_ = parameters_interface_->add_on_set_parameters_callback(update_param_cb);
      clock_ = rclcpp::Clock();
    }

    Params get_params() const{
      std::lock_guard<std::mutex> lock(mutex_);
      return params_;
    }

    bool is_old(Params const& other) const {
      std::lock_guard<std::mutex> lock(mutex_);
      return params_.__stamp != other.__stamp;
    }

    StackParams get_stack_params() {
      std::lock_guard<std::mutex> lock(mutex_);
      StackParams output;
      output.scientific_notation_num = params_.scientific_notation_num;
      output.pid.rate = params_.pid.rate;
      output.fixed_string = params_.fixed_string;
      output.fixed_array = params_.fixed_array;
      output.fixed_string_no_default = params_.fixed_string_no_default;
      output.kinematics.alpha = params_.kinematics.alpha;
      output.ft_sensor.frame.external = params_.ft_sensor.frame.external;
      output.ft_sensor.filter_coefficient = params_.ft_sensor.filter_coefficient;
      output.control.frame.external = params_.control.frame.external;
      output.fixed_world_frame.frame.external = params_.fixed_world_frame.frame.external;
      output.gravity_compensation.frame.external = params_.gravity_compensation.frame.external;
      output.gravity_compensation.CoG.force = params_.gravity_compensation.CoG.force;
      output.enable_parameter_update_without_reactivation = params_.enable_parameter_update_without_reactivation;
      output.use_feedforward_commanded_input = params_.use_feedforward_commanded_input;

      return output;
    }

    void refresh_dynamic_parameters() {
      std::lock_guard<std::mutex> lock(mutex_);
      // TODO remove any destroyed dynamic parameters

      // declare any new dynamic parameters
      rclcpp::Parameter param;
      for (const auto & value : params_.joints){
          auto entry = params_.pid.joints_map[value];
          auto param_name = fmt::format("{}.{}.{}", "pid", value, "p");
          if (!parameters_interface_->has_parameter(param_name)) {
              rcl_interfaces::msg::ParameterDescriptor descriptor;
              descriptor.description = "proportional gain term";
              descriptor.read_only = false;
              auto parameter = rclcpp::ParameterValue(entry.p);
              parameters_interface_->declare_parameter(param_name, parameter, descriptor);
          }
          param = parameters_interface_->get_parameter(param_name);
          if(auto validation_result = parameter_traits::lower_bounds<double>(param, 0);
            !validation_result.success()) {
              throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'pid.__map_joints.p': " + validation_result.error_msg()));
          }
          entry.p = param.as_double();}
      for (const auto & value : params_.joints){
          auto entry = params_.pid.joints_map[value];
          auto param_name = fmt::format("{}.{}.{}", "pid", value, "i");
          if (!parameters_interface_->has_parameter(param_name)) {
              rcl_interfaces::msg::ParameterDescriptor descriptor;
              descriptor.description = "integral gain term";
              descriptor.read_only = false;
              auto parameter = rclcpp::ParameterType::PARAMETER_DOUBLE;
              parameters_interface_->declare_parameter(param_name, parameter, descriptor);
          }
          param = parameters_interface_->get_parameter(param_name);
          entry.i = param.as_double();}
      for (const auto & value : params_.joints){
          auto entry = params_.pid.joints_map[value];
          auto param_name = fmt::format("{}.{}.{}", "pid", value, "d");
          if (!parameters_interface_->has_parameter(param_name)) {
              rcl_interfaces::msg::ParameterDescriptor descriptor;
              descriptor.description = "derivative gain term";
              descriptor.read_only = false;
              auto parameter = rclcpp::ParameterValue(entry.d);
              parameters_interface_->declare_parameter(param_name, parameter, descriptor);
          }
          param = parameters_interface_->get_parameter(param_name);
          entry.d = param.as_double();}
    }

    rcl_interfaces::msg::SetParametersResult update(const std::vector<rclcpp::Parameter> &parameters) {
      Params updated_params;
      {
        // Copy params_ so we only update it if all validation succeeds
        std::lock_guard<std::mutex> lock(mutex_);
        updated_params = params_;
      }

      for (const auto &param: parameters) {
        if (param.get_name() == "scientific_notation_num") {
            updated_params.scientific_notation_num = param.as_double();
        }
        if (param.get_name() == "interpolation_mode") {
            if(auto validation_result = parameter_traits::one_of<std::string>(param, {"spline", "linear"});
              !validation_result.success()) {
                return validation_result;
            }
            if(auto validation_result = parameter_traits::no_args_validator(param);
              !validation_result.success()) {
                return validation_result;
            }
            updated_params.interpolation_mode = param.as_string();
        }
        if (param.get_name() == "joints") {
            updated_params.joints = param.as_string_array();
        }
        if (param.get_name() == "pid.rate") {
            updated_params.pid.rate = param.as_double();
        }
        if (param.get_name() == "fixed_string") {
            if(auto validation_result = parameter_traits::size_lt<std::string>(param, 26);
              !validation_result.success()) {
                return validation_result;
            }
            updated_params.fixed_string = param.as_string();
        }
        if (param.get_name() == "fixed_array") {
            if(auto validation_result = parameter_traits::size_lt<double>(param, 11);
              !validation_result.success()) {
                return validation_result;
            }
            updated_params.fixed_array = param.as_double_array();
        }
        if (param.get_name() == "fixed_string_no_default") {
            if(auto validation_result = parameter_traits::size_lt<std::string>(param, 26);
              !validation_result.success()) {
                return validation_result;
            }
            updated_params.fixed_string_no_default = param.as_string();
        }
        if (param.get_name() == "command_interfaces") {
            updated_params.command_interfaces = param.as_string_array();
        }
        if (param.get_name() == "state_interfaces") {
            updated_params.state_interfaces = param.as_string_array();
        }
        if (param.get_name() == "chainable_command_interfaces") {
            updated_params.chainable_command_interfaces = param.as_string_array();
        }
        if (param.get_name() == "kinematics.plugin_name") {
            updated_params.kinematics.plugin_name = param.as_string();
        }
        if (param.get_name() == "kinematics.plugin_package") {
            updated_params.kinematics.plugin_package = param.as_string();
        }
        if (param.get_name() == "kinematics.base") {
            updated_params.kinematics.base = param.as_string();
        }
        if (param.get_name() == "kinematics.tip") {
            updated_params.kinematics.tip = param.as_string();
        }
        if (param.get_name() == "kinematics.alpha") {
            updated_params.kinematics.alpha = param.as_double();
        }
        if (param.get_name() == "kinematics.group_name") {
            updated_params.kinematics.group_name = param.as_string();
        }
        if (param.get_name() == "ft_sensor.name") {
            updated_params.ft_sensor.name = param.as_string();
        }
        if (param.get_name() == "ft_sensor.frame.id") {
            updated_params.ft_sensor.frame.id = param.as_string();
        }
        if (param.get_name() == "ft_sensor.frame.external") {
            updated_params.ft_sensor.frame.external = param.as_bool();
        }
        if (param.get_name() == "ft_sensor.filter_coefficient") {
            updated_params.ft_sensor.filter_coefficient = param.as_double();
        }
        if (param.get_name() == "control.frame.id") {
            updated_params.control.frame.id = param.as_string();
        }
        if (param.get_name() == "control.frame.external") {
            updated_params.control.frame.external = param.as_bool();
        }
        if (param.get_name() == "fixed_world_frame.frame.id") {
            updated_params.fixed_world_frame.frame.id = param.as_string();
        }
        if (param.get_name() == "fixed_world_frame.frame.external") {
            updated_params.fixed_world_frame.frame.external = param.as_bool();
        }
        if (param.get_name() == "gravity_compensation.frame.id") {
            updated_params.gravity_compensation.frame.id = param.as_string();
        }
        if (param.get_name() == "gravity_compensation.frame.external") {
            updated_params.gravity_compensation.frame.external = param.as_bool();
        }
        if (param.get_name() == "gravity_compensation.CoG.pos") {
            if(auto validation_result = parameter_traits::fixed_size<double>(param, 3);
              !validation_result.success()) {
                return validation_result;
            }
            updated_params.gravity_compensation.CoG.pos = param.as_double_array();
        }
        if (param.get_name() == "gravity_compensation.CoG.force") {
            updated_params.gravity_compensation.CoG.force = param.as_double();
        }
        if (param.get_name() == "admittance.selected_axes") {
            if(auto validation_result = parameter_traits::fixed_size<bool>(param, 6);
              !validation_result.success()) {
                return validation_result;
            }
            updated_params.admittance.selected_axes = param.as_bool_array();
        }
        if (param.get_name() == "admittance.mass") {
            if(auto validation_result = parameter_traits::fixed_size<double>(param, 6);
              !validation_result.success()) {
                return validation_result;
            }
            if(auto validation_result = parameter_traits::element_bounds<double>(param, 0.0001, 1000000.0);
              !validation_result.success()) {
                return validation_result;
            }
            updated_params.admittance.mass = param.as_double_array();
        }
        if (param.get_name() == "admittance.damping_ratio") {
            if(auto validation_result = parameter_traits::fixed_size<double>(param, 6);
              !validation_result.success()) {
                return validation_result;
            }
            if(auto validation_result = parameter_traits::validate_double_array_custom_func(param, 20.3, 5.0);
              !validation_result.success()) {
                return validation_result;
            }
            if(auto validation_result = parameter_traits::element_bounds<double>(param, 0.1, 10.0);
              !validation_result.success()) {
                return validation_result;
            }
            updated_params.admittance.damping_ratio = param.as_double_array();
        }
        if (param.get_name() == "admittance.stiffness") {
            if(auto validation_result = parameter_traits::element_bounds(param, 0.0001, 100000.0);
              !validation_result.success()) {
                return validation_result;
            }
            updated_params.admittance.stiffness = param.as_double_array();
        }
        if (param.get_name() == "enable_parameter_update_without_reactivation") {
            updated_params.enable_parameter_update_without_reactivation = param.as_bool();
        }
        if (param.get_name() == "use_feedforward_commanded_input") {
            updated_params.use_feedforward_commanded_input = param.as_bool();
        }
      }

      // update dynamic parameters
      for (const auto &param: parameters) {
        for (const auto & value : updated_params.joints){
            auto param_name = fmt::format("{}.{}.{}", "pid", value, "p");
            if (param.get_name() == param_name) {
                if(auto validation_result = parameter_traits::lower_bounds<double>(param, 0);
                  !validation_result.success()) {
                    return validation_result;
                }
                updated_params.pid.joints_map[value].p = param.as_double();
            }
        }
        for (const auto & value : updated_params.joints){
            auto param_name = fmt::format("{}.{}.{}", "pid", value, "i");
            if (param.get_name() == param_name) {
                updated_params.pid.joints_map[value].i = param.as_double();
            }
        }
        for (const auto & value : updated_params.joints){
            auto param_name = fmt::format("{}.{}.{}", "pid", value, "d");
            if (param.get_name() == param_name) {
                updated_params.pid.joints_map[value].d = param.as_double();
            }
        }
      }
      updated_params.__stamp = clock_.now();
      {
        std::lock_guard<std::mutex> lock(mutex_);
        params_ = updated_params;
      }
      return parameter_traits::OK;
    }

    void declare_params(){
      // declare all parameters and give default values to non-required ones
      if (!parameters_interface_->has_parameter("scientific_notation_num")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "Test scientific notation";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.scientific_notation_num);
          parameters_interface_->declare_parameter("scientific_notation_num", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("interpolation_mode")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies which algorithm to use for interpolation.";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.interpolation_mode);
          parameters_interface_->declare_parameter("interpolation_mode", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("joints")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies which joints will be used by the controller";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.joints);
          parameters_interface_->declare_parameter("joints", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("pid.rate")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "update loop period in seconds";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.pid.rate);
          parameters_interface_->declare_parameter("pid.rate", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("fixed_string")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "test code generation for fixed sized string";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.fixed_string);
          parameters_interface_->declare_parameter("fixed_string", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("fixed_array")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "test code generation for fixed sized array";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.fixed_array);
          parameters_interface_->declare_parameter("fixed_array", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("fixed_string_no_default")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "test code generation for fixed sized string with no default";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("fixed_string_no_default", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("command_interfaces")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies which command interfaces to claim";
          descriptor.read_only = true;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING_ARRAY;
          parameters_interface_->declare_parameter("command_interfaces", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("state_interfaces")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies which state interfaces to claim";
          descriptor.read_only = true;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING_ARRAY;
          parameters_interface_->declare_parameter("state_interfaces", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("chainable_command_interfaces")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies which chainable interfaces to claim";
          descriptor.read_only = true;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING_ARRAY;
          parameters_interface_->declare_parameter("chainable_command_interfaces", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("kinematics.plugin_name")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies which kinematics plugin to load";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("kinematics.plugin_name", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("kinematics.plugin_package")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies the package to load the kinematics plugin from";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("kinematics.plugin_package", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("kinematics.base")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies the base link of the robot description used by the kinematics plugin";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("kinematics.base", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("kinematics.tip")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies the end effector link of the robot description used by the kinematics plugin";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("kinematics.tip", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("kinematics.alpha")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies the damping coefficient for the Jacobian pseudo inverse";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.kinematics.alpha);
          parameters_interface_->declare_parameter("kinematics.alpha", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("kinematics.group_name")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies the group name for planning with Moveit";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("kinematics.group_name", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("ft_sensor.name")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "name of the force torque sensor in the robot description";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("ft_sensor.name", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("ft_sensor.frame.id")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "frame of the force torque sensor";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("ft_sensor.frame.id", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("ft_sensor.frame.external")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies if the force torque sensor is contained in the kinematics chain from the base to the tip";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.ft_sensor.frame.external);
          parameters_interface_->declare_parameter("ft_sensor.frame.external", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("ft_sensor.filter_coefficient")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies the coefficient for the sensor's exponential filter";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.ft_sensor.filter_coefficient);
          parameters_interface_->declare_parameter("ft_sensor.filter_coefficient", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("control.frame.id")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "control frame used for admittance control";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("control.frame.id", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("control.frame.external")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies if the control frame is contained in the kinematics chain from the base to the tip";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.control.frame.external);
          parameters_interface_->declare_parameter("control.frame.external", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("fixed_world_frame.frame.id")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "world frame, gravity points down (neg. Z) in this frame";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("fixed_world_frame.frame.id", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("fixed_world_frame.frame.external")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies if the world frame is contained in the kinematics chain from the base to the tip";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.fixed_world_frame.frame.external);
          parameters_interface_->declare_parameter("fixed_world_frame.frame.external", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("gravity_compensation.frame.id")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "frame which center of gravity (CoG) is defined in";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_STRING;
          parameters_interface_->declare_parameter("gravity_compensation.frame.id", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("gravity_compensation.frame.external")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies if the center of gravity frame is contained in the kinematics chain from the base to the tip";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.gravity_compensation.frame.external);
          parameters_interface_->declare_parameter("gravity_compensation.frame.external", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("gravity_compensation.CoG.pos")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "position of the center of gravity (CoG) in its frame";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_DOUBLE_ARRAY;
          parameters_interface_->declare_parameter("gravity_compensation.CoG.pos", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("gravity_compensation.CoG.force")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "weight of the end effector, e.g mass * 9.81";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.gravity_compensation.CoG.force);
          parameters_interface_->declare_parameter("gravity_compensation.CoG.force", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("admittance.selected_axes")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies if the axes x, y, z, rx, ry, and rz are enabled";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_BOOL_ARRAY;
          parameters_interface_->declare_parameter("admittance.selected_axes", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("admittance.mass")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies mass values for x, y, z, rx, ry, and rz used in the admittance calculation";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_DOUBLE_ARRAY;
          parameters_interface_->declare_parameter("admittance.mass", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("admittance.damping_ratio")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies damping ratio values for x, y, z, rx, ry, and rz used in the admittance calculation. The values are calculated as damping can be used instead: zeta = D / (2 * sqrt( M * S ))";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_DOUBLE_ARRAY;
          parameters_interface_->declare_parameter("admittance.damping_ratio", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("admittance.stiffness")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "specifies stiffness values for x, y, z, rx, ry, and rz used in the admittance calculation";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterType::PARAMETER_DOUBLE_ARRAY;
          parameters_interface_->declare_parameter("admittance.stiffness", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("enable_parameter_update_without_reactivation")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "if enabled, read_only parameters will be dynamically updated in the control loop";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.enable_parameter_update_without_reactivation);
          parameters_interface_->declare_parameter("enable_parameter_update_without_reactivation", parameter, descriptor);
      }
      if (!parameters_interface_->has_parameter("use_feedforward_commanded_input")) {
          rcl_interfaces::msg::ParameterDescriptor descriptor;
          descriptor.description = "if enabled, the velocity commanded to the admittance controller is added to its calculated admittance velocity";
          descriptor.read_only = false;
          auto parameter = rclcpp::ParameterValue(params_.use_feedforward_commanded_input);
          parameters_interface_->declare_parameter("use_feedforward_commanded_input", parameter, descriptor);
      }
      // get parameters and fill struct fields
      rclcpp::Parameter param;
      param = parameters_interface_->get_parameter("scientific_notation_num");
      params_.scientific_notation_num = param.as_double();
      param = parameters_interface_->get_parameter("interpolation_mode");
      if(auto validation_result = parameter_traits::one_of<std::string>(param, {"spline", "linear"});
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'interpolation_mode': " + validation_result.error_msg()));
      }
      if(auto validation_result = parameter_traits::no_args_validator(param);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'interpolation_mode': " + validation_result.error_msg()));
      }
      params_.interpolation_mode = param.as_string();
      param = parameters_interface_->get_parameter("joints");
      params_.joints = param.as_string_array();
      param = parameters_interface_->get_parameter("pid.rate");
      params_.pid.rate = param.as_double();
      param = parameters_interface_->get_parameter("fixed_string");
      if(auto validation_result = parameter_traits::size_lt<std::string>(param, 26);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'fixed_string': " + validation_result.error_msg()));
      }
      params_.fixed_string = param.as_string();
      param = parameters_interface_->get_parameter("fixed_array");
      if(auto validation_result = parameter_traits::size_lt<double>(param, 11);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'fixed_array': " + validation_result.error_msg()));
      }
      params_.fixed_array = param.as_double_array();
      param = parameters_interface_->get_parameter("fixed_string_no_default");
      if(auto validation_result = parameter_traits::size_lt<std::string>(param, 26);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'fixed_string_no_default': " + validation_result.error_msg()));
      }
      params_.fixed_string_no_default = param.as_string();
      param = parameters_interface_->get_parameter("command_interfaces");
      params_.command_interfaces = param.as_string_array();
      param = parameters_interface_->get_parameter("state_interfaces");
      params_.state_interfaces = param.as_string_array();
      param = parameters_interface_->get_parameter("chainable_command_interfaces");
      params_.chainable_command_interfaces = param.as_string_array();
      param = parameters_interface_->get_parameter("kinematics.plugin_name");
      params_.kinematics.plugin_name = param.as_string();
      param = parameters_interface_->get_parameter("kinematics.plugin_package");
      params_.kinematics.plugin_package = param.as_string();
      param = parameters_interface_->get_parameter("kinematics.base");
      params_.kinematics.base = param.as_string();
      param = parameters_interface_->get_parameter("kinematics.tip");
      params_.kinematics.tip = param.as_string();
      param = parameters_interface_->get_parameter("kinematics.alpha");
      params_.kinematics.alpha = param.as_double();
      param = parameters_interface_->get_parameter("kinematics.group_name");
      params_.kinematics.group_name = param.as_string();
      param = parameters_interface_->get_parameter("ft_sensor.name");
      params_.ft_sensor.name = param.as_string();
      param = parameters_interface_->get_parameter("ft_sensor.frame.id");
      params_.ft_sensor.frame.id = param.as_string();
      param = parameters_interface_->get_parameter("ft_sensor.frame.external");
      params_.ft_sensor.frame.external = param.as_bool();
      param = parameters_interface_->get_parameter("ft_sensor.filter_coefficient");
      params_.ft_sensor.filter_coefficient = param.as_double();
      param = parameters_interface_->get_parameter("control.frame.id");
      params_.control.frame.id = param.as_string();
      param = parameters_interface_->get_parameter("control.frame.external");
      params_.control.frame.external = param.as_bool();
      param = parameters_interface_->get_parameter("fixed_world_frame.frame.id");
      params_.fixed_world_frame.frame.id = param.as_string();
      param = parameters_interface_->get_parameter("fixed_world_frame.frame.external");
      params_.fixed_world_frame.frame.external = param.as_bool();
      param = parameters_interface_->get_parameter("gravity_compensation.frame.id");
      params_.gravity_compensation.frame.id = param.as_string();
      param = parameters_interface_->get_parameter("gravity_compensation.frame.external");
      params_.gravity_compensation.frame.external = param.as_bool();
      param = parameters_interface_->get_parameter("gravity_compensation.CoG.pos");
      if(auto validation_result = parameter_traits::fixed_size<double>(param, 3);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'gravity_compensation.CoG.pos': " + validation_result.error_msg()));
      }
      params_.gravity_compensation.CoG.pos = param.as_double_array();
      param = parameters_interface_->get_parameter("gravity_compensation.CoG.force");
      params_.gravity_compensation.CoG.force = param.as_double();
      param = parameters_interface_->get_parameter("admittance.selected_axes");
      if(auto validation_result = parameter_traits::fixed_size<bool>(param, 6);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'admittance.selected_axes': " + validation_result.error_msg()));
      }
      params_.admittance.selected_axes = param.as_bool_array();
      param = parameters_interface_->get_parameter("admittance.mass");
      if(auto validation_result = parameter_traits::fixed_size<double>(param, 6);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'admittance.mass': " + validation_result.error_msg()));
      }
      if(auto validation_result = parameter_traits::element_bounds<double>(param, 0.0001, 1000000.0);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'admittance.mass': " + validation_result.error_msg()));
      }
      params_.admittance.mass = param.as_double_array();
      param = parameters_interface_->get_parameter("admittance.damping_ratio");
      if(auto validation_result = parameter_traits::fixed_size<double>(param, 6);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'admittance.damping_ratio': " + validation_result.error_msg()));
      }
      if(auto validation_result = parameter_traits::validate_double_array_custom_func(param, 20.3, 5.0);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'admittance.damping_ratio': " + validation_result.error_msg()));
      }
      if(auto validation_result = parameter_traits::element_bounds<double>(param, 0.1, 10.0);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'admittance.damping_ratio': " + validation_result.error_msg()));
      }
      params_.admittance.damping_ratio = param.as_double_array();
      param = parameters_interface_->get_parameter("admittance.stiffness");
      if(auto validation_result = parameter_traits::element_bounds(param, 0.0001, 100000.0);
        !validation_result.success()) {
          throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'admittance.stiffness': " + validation_result.error_msg()));
      }
      params_.admittance.stiffness = param.as_double_array();
      param = parameters_interface_->get_parameter("enable_parameter_update_without_reactivation");
      params_.enable_parameter_update_without_reactivation = param.as_bool();
      param = parameters_interface_->get_parameter("use_feedforward_commanded_input");
      params_.use_feedforward_commanded_input = param.as_bool();


      // declare and set all dynamic parameters
      for (const auto & value : params_.joints){
          auto entry = params_.pid.joints_map[value];
          auto param_name = fmt::format("{}.{}.{}", "pid", value, "p");
          if (!parameters_interface_->has_parameter(param_name)) {
              rcl_interfaces::msg::ParameterDescriptor descriptor;
              descriptor.description = "proportional gain term";
              descriptor.read_only = false;
              auto parameter = rclcpp::ParameterValue(entry.p);
              parameters_interface_->declare_parameter(param_name, parameter, descriptor);
          }
          param = parameters_interface_->get_parameter(param_name);
          if(auto validation_result = parameter_traits::lower_bounds<double>(param, 0);
            !validation_result.success()) {
              throw rclcpp::exceptions::InvalidParameterValueException(fmt::format("Invalid value set during initialization for parameter 'pid.__map_joints.p': " + validation_result.error_msg()));
          }
          entry.p = param.as_double();}
      for (const auto & value : params_.joints){
          auto entry = params_.pid.joints_map[value];
          auto param_name = fmt::format("{}.{}.{}", "pid", value, "i");
          if (!parameters_interface_->has_parameter(param_name)) {
              rcl_interfaces::msg::ParameterDescriptor descriptor;
              descriptor.description = "integral gain term";
              descriptor.read_only = false;
              auto parameter = rclcpp::ParameterType::PARAMETER_DOUBLE;
              parameters_interface_->declare_parameter(param_name, parameter, descriptor);
          }
          param = parameters_interface_->get_parameter(param_name);
          entry.i = param.as_double();}
      for (const auto & value : params_.joints){
          auto entry = params_.pid.joints_map[value];
          auto param_name = fmt::format("{}.{}.{}", "pid", value, "d");
          if (!parameters_interface_->has_parameter(param_name)) {
              rcl_interfaces::msg::ParameterDescriptor descriptor;
              descriptor.description = "derivative gain term";
              descriptor.read_only = false;
              auto parameter = rclcpp::ParameterValue(entry.d);
              parameters_interface_->declare_parameter(param_name, parameter, descriptor);
          }
          param = parameters_interface_->get_parameter(param_name);
          entry.d = param.as_double();}

      params_.__stamp = clock_.now();
    }

    private:
      Params params_;
      rclcpp::Clock clock_;
      std::shared_ptr<rclcpp::node_interfaces::OnSetParametersCallbackHandle> handle_;
      std::shared_ptr<rclcpp::node_interfaces::NodeParametersInterface> parameters_interface_;
      std::mutex mutable mutex_;
  };

} // namespace admittance_controller